package client;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import simulation.Constants;
import tracker.JoinReply;
import tracker.PeerReply;
import tracker.TCPTrackerClient;
import tracker.TrackerClient;
import tracker.UDPTrackerClient;
import util.File;
import util.RandomUtil;
import util.Serializer;
import util.ThreadUtil;
import util.File.Piece;
import client.PeerState.Neighbor;
import client.messages.BitMessage;
import client.messages.ChunkReply;
import client.messages.ChunkRequest;
import client.messages.ForwardMessage;
import client.messages.HaveMessage;
import client.messages.Message;
import client.messages.MessageType;

import common.Dispatcher;
import common.Pair;
import common.PeerAddress;
import common.Server;

/**
 * The Peer class abstracts away socket communication between peers.
 * 
 * @author kuat
 * @author albertni
 */
public final class Peer implements Dispatcher, Runnable {
	private final Server server;
	private final PeerState state;
	private final TrackerClient trackerClient;

	private final Map<Integer, DownloadSession> completed = new HashMap<Integer, DownloadSession>();
	private final Map<Integer, DownloadSession> sessions = new HashMap<Integer, DownloadSession>();

	private final Statistics stat = new Statistics();

	private final Random random = new Random();

	/**
	 * Simply creates a Peer with just one initial file
	 * 
	 * @param port
	 * @param trackerPort
	 * @param initialFile
	 */
	public Peer(int port, String tracker, int trackerPort, File initialFile) {
		this(port, tracker, trackerPort, new File[] { initialFile });
	}

	// non-blocking
	public Peer(int port, String tracker, int trackerPort, File[] initialFiles) {
		this.state = new PeerState(port);

		for (File f : initialFiles)
			state.addFile(f);

		Timer.singleton.register(this);
		
		// start a tracker client
		try {
			trackerClient = Constants.USE_TCP ? new TCPTrackerClient(tracker,
					trackerPort) : new UDPTrackerClient(tracker, trackerPort);
		} catch (IOException e) {
			e.printStackTrace();
			System.err.println("could not create a torrent client");
			throw new Error();
		}

		// start a server
		try {
			server = new Server(port, this);
		} catch (IOException e) {
			e.printStackTrace();
			System.err.println("could not create a socket for port " + port);
			throw new Error();
		}
		
		new Thread(this).start();
	}

	/**
	 * Factory method for generating a peer that starts with a RANDOM SUBPART of
	 * f. Convenience method
	 * 
	 * @requires f is a FULL file
	 * @param port
	 * @param trackerPort
	 * @param f
	 * @return a peer that contains a random subpart of f
	 */
	public static Peer generatePeer(int port, String tracker, int trackerPort,
			File f) {
		return new Peer(port, tracker, trackerPort, new File[] { f
				.getRandomSubPart(Math.random()) });
	}

	/**
	 * Not thread-safe.
	 * 
	 * @param s
	 * @return
	 */
	private Neighbor neighbor(String s) {
		assert s != null;
		
		if (state.getNeighbor(s) == null) {
			Neighbor n = state.new Neighbor(s);
			state.addNeighbor(n);
			advertisePieces(n);
		}

		return state.getNeighbor(s);
	}

	// not thread-safe
	private void send(Neighbor n, Message s) {
		stat.sent();		

		if (Constants.FORWARDING) {
			ForwardMessage fm = new ForwardMessage();
			fm.data = s;
			fm.sourceName = state.name();
			fm.destinationName = n.name;

			if (n.forwarder == null) {
				PeerAddress a = null;
				try {
					a = trackerClient.circuit(state.name(), n.name);
				} catch (IOException e) {
					e.printStackTrace();
				}

				assert a != null : "no circuit";
				print("connected to " + s + " through circuit " + a);

				n.forwarder = a;
			}

			assert n.address == null;
			server.send(n.forwarder, fm);
		} else {
			s.sourceName = state.name();
			if (n.address == null) {
				try {
					assert n.name != null;
					n.address = trackerClient.ip(n.name);
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
			server.send(n.address, s);
		}
	}

	// not thread-safe
	private void advertisePieces(Neighbor s) {
		Set<Integer> indexes = new HashSet<Integer>();
		File file = state.file();

		Set<Integer> incomplete = file.incomplete();
		for (int i = 0; i < file.numberPieces(); i++)
			if (!incomplete.contains(i))
				indexes.add(i);

		Message m = new HaveMessage(indexes);
		m.sourceName = state.name();
		send(s, m);
	}

	class RateComparator implements Comparator<Pair<Double, Neighbor>> {
		@Override
		public int compare(Pair<Double, Neighbor> arg0,
				Pair<Double, Neighbor> arg1) {
			double d1 = arg0.first();
			double d2 = arg1.first();
			if (reverse)
				return Double.compare(d2, d1);
			else
				return Double.compare(d1, d2);
		}

		boolean reverse;

		public RateComparator(boolean reverse) {
			this.reverse = reverse;
		}
	}

	private void updateChoking2() {

		Set<Neighbor> downloaders = new HashSet<Neighbor>();
		Set<Neighbor> niceGuys = new HashSet<Neighbor>();
		Neighbor worstDownloader = null;
		
		List<Pair<Double, Neighbor>> rateList = new ArrayList<Pair<Double, Neighbor>>(
				state.getNeighbors().size());
		for (Neighbor n : state.getNeighbors())
			rateList.add(new Pair<Double, Neighbor>(n.upload.measure(), n));
		Collections.sort(rateList, new RateComparator(true));
		// rateList has the upload rates for everyone, sorted from highest to
		// lowest
		Set<Neighbor> willBeChoked = new HashSet<Neighbor>();
		int numDownloaders = 0;
		Iterator<Pair<Double, Neighbor>> i = rateList.iterator();
		while (i.hasNext()) {
			Pair<Double, Neighbor> p = i.next();
			Neighbor n = p.second();
			if (numDownloaders >= 4) {
				willBeChoked.add(n);
				continue;
			} // we aren't going to talk to anyone else here, have enough
				// downloaders
			if (n.peer_interested) {
				downloaders.add(n);
				numDownloaders++;
				worstDownloader = n;
			} else {
				// peer isn't interested
				// unchoke anyway if upload bandwidth is comparable to the
				// downloaders
				niceGuys.add(n);
			}
		}
		if (willBeChoked.size() > 0) {
			Neighbor optimisticUnchoke = RandomUtil.getRandom(willBeChoked);
			willBeChoked.remove(optimisticUnchoke);
			if (optimisticUnchoke.peer_interested) {
				downloaders.add(optimisticUnchoke);
				downloaders.remove(worstDownloader);
				willBeChoked.add(worstDownloader);
			} else
				niceGuys.add(optimisticUnchoke);
		}
		for (Neighbor n : downloaders)
			setAmChoking(n, false);
		for (Neighbor n : niceGuys)
			setAmChoking(n, false);
		for (Neighbor n : willBeChoked)
			setAmChoking(n, true);
	}

/*	private void updateChoking() {
		// http://wiki.theory.org/BitTorrentSpecification#Choking_and_Optimistic_Unchoking
		Set<Neighbor> unchokeCandidates = new HashSet<Neighbor>(state
				.getNeighbors());
		List<Pair<Double, Neighbor>> rateList = new ArrayList<Pair<Double, Neighbor>>(
				unchokeCandidates.size());
		for (Neighbor n : unchokeCandidates)
		rateList.add(new Pair<Double,Neighbor>(n.upload.measure(), n));
		Collections.sort(rateList, new RateComparator(true));

		Set<Neighbor> willBeUnchoked = new HashSet<Neighbor>();
		for (int i = 0; i < Math.min(unchokeCandidates.size(), 4); i++)
			willBeUnchoked.add(rateList.get(i).second());

		Set<Neighbor> willBeChoked = new HashSet<Neighbor>();
		for (int i = Math.min(unchokeCandidates.size(), 4); i < unchokeCandidates
				.size(); i++) {
			// people with poor upload rates.
			willBeChoked.add(rateList.get(i).second());
			// the only exception are people who arent interested but still have
			// good upload rate
		}

		// 2. optimistically unchoke one peer (if he is one of the downloaders
		// don't bother). FIXME: should cycle through which one gets unchoked
		if (willBeChoked.size() > 0) {
			Neighbor optimisticUnchoke = RandomUtil.getRandom(willBeChoked);
			willBeChoked.remove(optimisticUnchoke);
			willBeUnchoked.add(optimisticUnchoke);
		}

		for (Neighbor n : willBeUnchoked)
			setAmChoking(n, false);
		for (Neighbor n : willBeChoked)
			setAmChoking(n, true);
	}*/

	private Integer updateInterests() {

		// pieces neighbors have
		Set<Integer> owned = state.getNeighborPieces();
		// pieces I don't have
		Set<Integer> needed = state.file().incomplete();

		// Update interested status
		Collection<Neighbor> losers = new ArrayList<Neighbor>(state
				.getNeighbors());
		for (Neighbor n : state.getNeighbors()) {
			// if he has a piece I need, he is not a loser
			for (Integer piece : n.pieces()) {
				if (needed.contains(piece)) {
					losers.remove(n);
					setAmInterested(n, true);
					break;
				}
			}
		}

		for (Neighbor n : losers)
			setAmInterested(n, false);

		// pieces I don't have, I need, and not downloading at the moment
		List<Integer> desired = new ArrayList<Integer>();
		for (Integer piece : needed) {
			if (owned.contains(piece) && !sessions.keySet().contains(piece))
				desired.add(piece);

		}

		if (desired.size() > 0) {
			Integer piece = desired.get(random.nextInt(desired.size()));

			Collection<Neighbor> owners = state.getNeighborForPiece(piece);

			// someone has it?
			assert owners.size() > 0;

			// not requested?
			assert !sessions.keySet().contains(piece);
			assert !completed.keySet().contains(piece);

			// interested?
			for (Neighbor n : owners)
				assert n.am_interested;
			return piece;
		} else {

			return null;
		}
	}

	private Neighbor selectNeighbor(Integer piece) {
		Collection<Neighbor> owners = state.getNeighborForPiece(piece);

		// FIXME: need to chose person who download from better
		// Choose 4 persons who arent choking you and have good
		// upload speed

		for (Neighbor owner : owners)
			if (!owner.peer_choking && owner.am_interested) {
				return owner;
			}

//		print("can't discover any good neighbor for piece " + piece);
		return null;
	}

	public void run() {
		// first call to the tracker
		synchronized (state) {
			try {
				JoinReply reply = trackerClient.connect(state.file().name(),
						state.port(), -1);
				state.setName(reply.name);
				Thread.currentThread().setName(
						"Main loop of peer " + reply.name + " @ "
								+ state.port());

				ThreadUtil.randomSleep(100, 900);
				
				PeerReply peers = trackerClient.peers(reply.name, state.file()
						.name());

				for (String s : peers.peers)
					advertisePieces(neighbor(s));

			} catch (IOException e) {
				e.printStackTrace();
			}
		}

		Task readvertise = new Task(30000, new Runnable() {
			public void run() {
				synchronized (state) {
					for (Neighbor n : state.getNeighbors())
						advertisePieces(n);
				}
			}
		});

		Task updateChoking = new Task(10000, new Runnable() {
			public void run() {
				synchronized (state) {
					updateChoking2();
				}
			}
		});

		Task staleDownloads = new Task(Constants.FAST_SESSION_TIMEOUT,
				new Runnable() {
					public void run() {
						synchronized (state) {
							for (DownloadSession session : sessions.values()) {
								if (session.stale()) {
									print("Stale download session discovered to "
											+ session.neighbor());
									Neighbor n = selectNeighbor(session.piece);
									if (n != null)
										send(n, session.start(n));
								}
							}
						}
					}
				});

		while (true) {
			readvertise.check();
			updateChoking.check();
			staleDownloads.check();

			synchronized (state) {
				File file = state.file();

				if (!file.isComplete()
						&& sessions.keySet().size() < Constants.PARALLEL_SESSIONS) {
					// owned and needed but not requested
					Integer desiredPiece = updateInterests();
					if (desiredPiece != null) {
						Neighbor owner = selectNeighbor(desiredPiece);
						if (owner != null) {
							DownloadSession session = new DownloadSession(state
									.name(), desiredPiece);
							sessions.put(desiredPiece, session);
							send(owner, session.start(owner));
						}
					}
				} 
					
				if (file.isComplete())
					Timer.singleton.finished(this);
			}

			ThreadUtil.sleep();
		}
	}

	/**
	 * Not thread-safe.
	 * 
	 * @param msg
	 */
	private final void print(String msg) {
		StringBuilder sb = new StringBuilder();
		sb.append("[").append(state.name()).append(" @ ").append(state.port())
				.append("] ").append(msg);
		System.out.println(sb);
	}

	// actual addresses XXX: remove it later on
	Map<String, PeerAddress> addresses = new HashMap<String, PeerAddress>();

	@Override
	/**
	 * Processes the given bytebuffer which contains message data
	 */
	public void process(byte[] bb, String host, int port) {
		Message m = (Message) Serializer.read(bb);
		// print("processing " + m + " from " + m.sourceName);

		synchronized (state) {
			String sourceName = ((Message) m).sourceName;
			assert sourceName != null;

			if (m instanceof HaveMessage) {
				stat.received();
				Set<Integer> indexes = ((HaveMessage) m).indexes;
				for (Integer index : indexes)
					state.registerNeighborPiece(index, neighbor(sourceName));
			} else if (m instanceof ChunkRequest) {
				stat.received();				

				Neighbor n = neighbor(sourceName);
				n.download.ensureStarted();
				if (n.peer_interested && !n.am_choking) {
					ChunkRequest cr = (ChunkRequest) m;
					File f = state.file();
					Piece relevantPiece = f.getPiece(cr.pieceIndex);

					for (int index : cr.chunkIndex) {
						String block = relevantPiece.get(index);
						assert block.length() == Constants.CHUNK_SIZE;

						ChunkReply reply = new ChunkReply();

						reply.pieceIndex = cr.pieceIndex;
						reply.chunkIndex = index;
						reply.block = block;
						reply.sourceName = this.state.name();

						n.download.report();
						send(n, reply);
					}
				}
			} else if (m instanceof BitMessage) {
				stat.received();

				Neighbor n = neighbor(sourceName);
				BitMessage b = (BitMessage) m;
				switch (b.type) {
				case INTERESTED:
					n.peer_interested = true;
					break;
				case NOT_INTERESTED:
					n.peer_interested = false;
					break;
				case CHOKE:
					n.peer_choking = true;
					break;
				case UNCHOKE:
					n.peer_choking = false;
					break;
				default:
					throw new RuntimeException(
							"Unknown type of BitMessage instance in Peer.process()!");
				}
			} else if (m instanceof ChunkReply) {
				stat.received();

				ChunkReply cr = (ChunkReply) m;
				File f = state.file();

				Piece piece = f.getPiece(cr.pieceIndex);
				assert piece != null;
				piece.put(cr.chunkIndex, cr.block);

				DownloadSession session = sessions.get(piece.index());
				// too late packet
				if (session != null) {

					// continue polling stuff
					ChunkRequest msg = session.proceed(cr);
					if (msg != null)
						send(session.neighbor(), msg);

					if (piece.isComplete()) {
						for (Neighbor x : state.getNeighbors())
							advertisePieces(x);

						// propagate from sessions to completed
						completed.put(piece.index(), session);
						sessions.remove(piece.index());
					}
				}

			} else if (m instanceof ForwardMessage) {
				stat.forwarded();
				ForwardMessage fm = (ForwardMessage) m;

				fm.data.sourceName = fm.sourceName;

				if (!addresses.containsKey(fm.destinationName))
					try {
						addresses.put(fm.destinationName, trackerClient
								.ip(fm.destinationName));
					} catch (IOException e) {
						e.printStackTrace();
					}

				PeerAddress destination = addresses.get(fm.destinationName);
				if (destination != null)
					server.send(destination, fm.data);
			} else
				assert false;
		}
	}

	private void setAmInterested(Neighbor n, boolean am_interested) {
		if (n.am_interested == am_interested) {
			return; // this is just a NOP
		}

		n.am_interested = am_interested;
		if (am_interested)
			send(n, new BitMessage(MessageType.INTERESTED));
		else
			send(n, new BitMessage(MessageType.NOT_INTERESTED));
	}

	private void setAmChoking(Neighbor n, boolean am_choking) {
		if (n.am_choking == am_choking)
			return; // this is just a NOP

		n.am_choking = am_choking;
		if (am_choking)
			send(n, new BitMessage(MessageType.CHOKE));
		else
			send(n, new BitMessage(MessageType.UNCHOKE));
	}

	PeerState state() {
		return state;
	}

	Statistics statistics() {
		return stat;
	}
}
