package client;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import util.File;

import common.PeerAddress;

/**
 * Client state, not per-file state:
 * <ol>
 * <li> name of the peer;
 * <li> files
 * <li> neighbors and their forwarders
 * </ol>
 * 
 * @author kuat
 * @author amrik
 * 
 */
final class PeerState {
	/** Name of this local peer */
	private String name;
	private int port;

	/** Set of files that this peer has or wants */
	private final Set<File> fileSet = new HashSet<File>();

	private final Map<String, Neighbor> neighbors = new HashMap<String, Neighbor>();

	/**
	 * @return a unmodifiable view of the set of files that this peer knows
	 *         about
	 */
	public Set<File> getFiles() {
		return Collections.unmodifiableSet(fileSet);
	}

	/**
	 * Adds a file to the set of files
	 */
	public boolean addFile(File f) {
		return fileSet.add(f);
	}

	/**
	 * Removes a file from the set of files
	 */
	public boolean removeFile(File f) {
		return fileSet.remove(f);
	}

	/**
	 * A <b>Neighbor</b> is a representation of a peer whose PeerAddress you
	 * know. An instance of this class stores important choking state
	 * information such as whether we are choking or being choked, and whether
	 * the remote peer is interested etc.
	 * 
	 * @author amrik
	 */
	final class Neighbor implements Comparable<Neighbor> {		
		public final String name;
		private final Set<Integer> pieces;
		/** upload statistics: how much he sends to me */
		public final Rate upload;
		/** download statistics: how much i sent to him */
		public final Rate download;

		public Neighbor(String name) {			
			this.name = name;
			this.pieces = new HashSet<Integer>();
			this.upload = new Rate();
			this.download = new Rate();
		}
		
		public Set<Integer> pieces() {
			return Collections.unmodifiableSet(pieces);
		}
		
		public void addPiece(int i) {
			pieces.add(i);
		}
		
		/** actual address; should not be used if forwarding */
		public volatile PeerAddress address;
		/** temporary forwarder */
		public volatile PeerAddress forwarder;		
		/** this client is choking the peer */
		public volatile boolean am_choking = true;
		/** this client is interested in the peer */
		public volatile boolean am_interested = false;
		/** peer is choking this client */
		public volatile boolean peer_choking = true;
		/** peer is interested in this client */
		public volatile boolean peer_interested = false;

		@Override
		public boolean equals(Object obj) {
			if (obj instanceof Neighbor) {
				Neighbor n = (Neighbor) obj;
				return (n.name.equals(this.name));
			} else
				return false;
		}

		@Override
		public int hashCode() {
			return name.hashCode();
		}
		
		@Override
		public String toString() {
			return name + "I" + (am_interested ? "1" : "0") + (peer_interested ? "1" : "0") + "C" + (am_choking ? "1" : "0") + (peer_choking ? "1" : "0") + "U" + upload.measure();
		}

		@Override
		public int compareTo(Neighbor o) {			
			return name.compareTo(o.name);
		}
	}

	protected PeerState(int localPort) {
		this.port = localPort;
	}

	/** @return a unmodifiable view of the set of neighbors of this peer */
	Collection<Neighbor> getNeighbors() {
		return Collections.unmodifiableCollection(neighbors.values());
	}

	Neighbor getNeighbor(String s) {
		return neighbors.get(s);
	}

	/**
	 * Adds the neighbor <tt>n</tt> to the set of neighbors. Beware: two
	 * neighbors are considered equal iff both their PeerAddresses and names are
	 * the same.
	 * 
	 * @param n
	 *            the neighbor to remove
	 * @return true if n was not already in the set of neighbors
	 */
	Neighbor addNeighbor(Neighbor n) {
		return neighbors.put(n.name, n);
	}

	/**
	 * Removes the neighbor <tt>n</tt> from the set of neighbors
	 * 
	 * @param n
	 *            the neighbor to remove
	 * @return true if n was in the set of neighbors
	 */
	Neighbor removeNeighbor(Neighbor n) {
		return neighbors.remove(n.name);
	}

	/** @return name of this Peer */
	public String name() {
		return name;
	}

	void setName(String name) {
		assert this.name == null && name != null;
		this.name = name;
	}

	public int port() {
		return port;
	}

	public File file() {
		return fileSet.iterator().next();
	}

	/** Map piece # to list of peers that have it */
	private Map<Integer, List<Neighbor>> pieces = new HashMap<Integer, List<Neighbor>>();

	/** @return Map piece # to list of peers that have it */
	void registerNeighborPiece(Integer p, Neighbor n) {
		if (!pieces.containsKey(p))
			pieces.put(p, new ArrayList<Neighbor>());
		pieces.get(p).add(n);
		n.addPiece(p);
		repOK();
	}

	Collection<Neighbor> getNeighborForPiece(Integer p) {
		if (!pieces.containsKey(p))
			return Collections.emptySet();
		else
			return Collections.unmodifiableCollection(pieces.get(p));
	}
	
	public Set<Integer> getNeighborPieces() {
		return Collections.unmodifiableSet(pieces.keySet());
	}
	
	private final void repOK() {
		for (Integer piece : pieces.keySet()) 
			for (Neighbor n : pieces.get(piece))
				assert n.pieces().contains(piece); 
		for (Neighbor n : neighbors.values())
			for (Integer piece : n.pieces())
				assert pieces.get(piece) != null && pieces.get(piece).contains(n);
	}
}
