package common;

import java.io.IOException;
import java.io.Serializable;
import java.net.InetSocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.ServerSocketChannel;
import java.nio.channels.SocketChannel;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import simulation.Constants;
import util.Serializer;
import util.ThreadUtil;
import client.DefaultLatencyEmulator;
import client.LatencyEmulator;

/**
 * The Server class abstracts a socket bound on a particular port. This listens
 * for incoming connections from other clients.
 * 
 */
public class Server {

	private static LatencyEmulator emulator = new DefaultLatencyEmulator();
	
	public static void setEmulator(LatencyEmulator e){
		emulator = e;
	}
	
	public static LatencyEmulator emulator() {
		return emulator;
	}
	
	private final ServerSocketChannel socket;
	private final Dispatcher dispatcher;
	private final List<SocketChannel> remote2local;
	private final Map<SocketChannel, IncompleteBuffer> incompleteBuffers;

	class IncompleteBuffer {
		public int length = 0;
		public ByteBuffer bb = ByteBuffer.allocate(BYTE_BUFFER_CAPACITY * 2);

		public IncompleteBuffer() {
			bb.position(0);
		}
	}

	private final List<Pair<PeerAddress, Serializable>> outgoing;

	/**
	 * I have no idea how big this is actually supposed to be, currently set to
	 * 17*1024 since the chunk size in BitTorrent is 16 KB and we need some
	 * extra bytes for additional information
	 */
	private static final int BYTE_BUFFER_CAPACITY = 1024;

	public Server(int port, Dispatcher dispatcher) throws IOException {
		assert dispatcher != null;
		assert port > 0;
		assert port < 65536;

		// create a server socket
		this.socket = ServerSocketChannel.open();
		this.socket.configureBlocking(true);
		this.socket.socket().bind(new InetSocketAddress(port));

		this.dispatcher = dispatcher;
		this.local2remote = new HashMap<PeerAddress, SocketChannel>();
		this.remote2local = new ArrayList<SocketChannel>();
		this.outgoing = new ArrayList<Pair<PeerAddress, Serializable>>();

		this.incompleteBuffers = new HashMap<SocketChannel, IncompleteBuffer>();

		new Thread(new Runnable() {
			public void run() {
				listen();
			}
		}).start();

		new Thread(new Runnable() {
			public void run() {
				process();
			}
		}).start();

		new Thread(new Runnable() {
			public void run() {
				send();
			}
		}).start();
	}

	private void listen() {
		print("Server Listening");
		Thread.currentThread().setName(
				"Listener @ " + socket.socket().getLocalPort());
		while (true) {
			try {
				SocketChannel n = socket.accept();
				n.configureBlocking(false);
				synchronized (remote2local) {
					if (remote2local.contains(n))
						n.close();
					else {
//						print("accepted a connection " + n);
						remote2local.add(n);
					}
				}

			} catch (IOException e1) {
				e1.printStackTrace();
			}

			ThreadUtil.sleep();
		}
	}

	private void process() {
		print("Started message dispatcher");
		Thread.currentThread().setName(
				"Dispatcher @ " + socket.socket().getLocalPort());
		while (true) {

			List<SocketChannel> peers;
			synchronized (remote2local) {
				peers = new ArrayList<SocketChannel>(remote2local);
			}

			for (SocketChannel sc : peers) {
				if (!sc.isOpen() || !sc.isConnected())
					continue;

				// allocate a temp buffer if it doesn't already exist
				if (!incompleteBuffers.containsKey(sc))
					incompleteBuffers.put(sc, new IncompleteBuffer());

				IncompleteBuffer ib = incompleteBuffers.get(sc);
				ByteBuffer bb = ib.bb;

				try {
					int k = sc.read(bb);
					if (k == -1)
						throw new IOException("Connection is closed");

					ib.length += k;

					byte[] msg = new byte[BYTE_BUFFER_CAPACITY];
					if (ib.length >= BYTE_BUFFER_CAPACITY) {
						bb.position(0);
						bb.get(msg);
						bb.position(BYTE_BUFFER_CAPACITY);
						bb.compact();
						ib.length -= BYTE_BUFFER_CAPACITY;
						bb.position(ib.length);

						dispatcher.process(msg, sc.socket().getInetAddress().getHostName(), sc.socket().getPort());
					}

				} catch (IOException e) {
					synchronized (remote2local) {
						remote2local.remove(sc);
					}

					try {
						sc.close();
					} catch (IOException e1) {
						e1.printStackTrace();
					}
				}
			}

			ThreadUtil.fastSleep();
		}
	}

	private final Map<PeerAddress, SocketChannel> local2remote;

	private void send() {
		print("Started message sender at ");
		Thread.currentThread().setName(
				"Sender @ " + socket.socket().getLocalPort());
		while (true) {
			Pair<PeerAddress, Serializable> next = null;

			synchronized (outgoing) {
				if (outgoing.size() > 0) {
					next = outgoing.remove(0);
				}
			}

			if (next != null) {
				PeerAddress addr = next.first();
				SocketChannel sc = local2remote.get(addr);

				if (sc == null) {
					try {
						sc = SocketChannel.open();
						sc.configureBlocking(false);
						sc.connect(new InetSocketAddress(addr.host(), addr
								.port()));
						while (!sc.finishConnect())
							ThreadUtil.sleep();
						local2remote.put(addr, sc);
					} catch (Exception e) {
						e.printStackTrace();
					}
				}

				assert sc != null : "no socket channel found or created";

				Serializable msg = next.second();
				byte[] ser = Serializer.write(msg);
				if (ser.length > BYTE_BUFFER_CAPACITY) {
					System.err.println("BUFFER OVERFLOW! Killing the JVM process");
					System.err.println(msg);
					System.err.println("Length: " + ser.length);
					throw new Error();
				}
				assert ser.length <= BYTE_BUFFER_CAPACITY : "buffer overflow for the message";
				ByteBuffer buf = ByteBuffer.allocate(BYTE_BUFFER_CAPACITY);

				for (int i = 0; i < ser.length; i++)
					buf.put(i, ser[i]);
				for (int i = ser.length; i < buf.capacity(); i++)
					buf.put(i, (byte) 0);

				try {
					int counter = ser.length;
					do {
						if (Constants.LATENCY) 
							ThreadUtil.sleep(emulator.sleep(socket.socket().getLocalPort(), addr.port()));
						int w = sc.write(buf);
						if (w == -1)
							throw new IOException(
									"writing to socket channel failed");
						counter -= w;
					} while (counter > 0);
				} catch (IOException e) {
					e.printStackTrace();
				}
			} else
				ThreadUtil.sleep();
		}
	}

	public void send(PeerAddress peer, Serializable msg) {
		synchronized (outgoing) {
			outgoing.add(new Pair<PeerAddress, Serializable>(peer, msg));
		}
	}

	private void print(String msg) {
		System.out.println("[server @ " + socket.socket().getLocalPort() + "] "
				+ msg);
	}	
}
