package tracker;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import simulation.Constants;
import util.RandomUtil;

import common.PeerAddress;

/**
 * The mutable state of the tracker. 
 * 
 * @specfield name2info : set Name -> PeerInfo
 * @specfield torrents: set FileId -> TorrentInfo
 * @author kuat
 *
 */
class TrackerState {
	
	/**
	 * Peer information.
	 * 
	 * @author kuat
	 *
	 */
	class PeerInfo {
		final String name;
		final PeerAddress addr;
		final PeerAddress udp;
		
		PeerInfo(String name, int port, int udp, String host) {
			this.name = name;
			this.addr = new PeerAddress(port, host);
			this.udp = new PeerAddress(udp, host);
		}
	}
	
	/**
	 * Torrent information.
	 * 
	 * @specfield peers: set Name 
	 * @author kuat
	 *
	 */
	class TorrentInfo {
		final String fileId;
		private final List<String> peers;
		
		TorrentInfo(String fileId) {
			this.fileId = fileId;
			this.peers = new ArrayList<String>();
		}
		
		List<String> peers() {
			return Collections.unmodifiableList(peers);
		}
		
		void addPeer(String name) {
			peers.add(name);
		}
	}

	
	private int counter;
		
	// Peer mappings
	private final Map<Integer, PeerInfo> port2info;	
	private final Map<String, PeerInfo> name2info;		
	
	// File mapping
	private final Map<String, TorrentInfo> torrents;
	
	TrackerState() {
		this.name2info = new HashMap<String, PeerInfo>();
		this.port2info = new HashMap<Integer, PeerInfo>();
		this.torrents = new HashMap<String, TorrentInfo>();
		this.counter = 0;
	}
	
	
	/**
	 * Assign a name to the guy on the port if it does not have one.
	 * @param tcp port
	 * @param udp port
	 * @return
	 */
	synchronized String register(int tcp, int udp, String host) {
		assert RepOK();		
		Integer i = new Integer(tcp);
		
		if (port2info.containsKey(i)) 
			return port2info.get(i).name;
				
		String name = "P" + counter++;
		System.out.println("[Tracker] registering " + name + " at " + tcp);
		PeerInfo info = new PeerInfo(name, tcp, udp, host);
		
		name2info.put(name, info);
		port2info.put(i, info);
		return name;
	}		
	
	/**
	 * Register the name as a download for the given file,
	 * @param name
	 * @param fileId
	 * @return
	 */
	synchronized void register(String name, String fileId) {
		// add a new torrent if not found
		if (!torrents.containsKey(fileId)) {
			torrents.put(fileId, new TorrentInfo(fileId));
		}
		
		TorrentInfo info = torrents.get(fileId);
		assert !info.peers().contains(name);
		info.addPeer(name);
	}
	
	synchronized Set<String> peers(String name, String fileId) {
		TorrentInfo info = torrents.get(fileId);
		List<String> peers = new ArrayList<String>(info.peers());
		assert peers.contains(name);
		peers.remove(name);
		
		Set<String> output = RandomUtil.getRandom(peers, Constants.NEIGHBOURS_REPLY);
		return output;
	}
	
	synchronized PeerInfo getAddress(String name) {
		return name2info.get(name);
	}
	
	synchronized PeerInfo getRandomAddress(String ... args) {
		Map<String, PeerInfo> peers = new HashMap<String, PeerInfo>(name2info);
		for (String s : args)
			peers.remove(s);
		return RandomUtil.getRandom(new ArrayList<PeerInfo>(peers.values()));
	}
		
	private final boolean RepOK(){
		assert port2info.keySet().size() == name2info.keySet().size();
		assert port2info.values().size() == port2info.values().size();
		return true;
	}
}