/**
 * 
 */
package util;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Random;
import java.util.Set;

import simulation.Constants;
import simulation.Manager;

/**
 * @author turalb, albertni, kuat
 * 
 */
public class File {

	private final Piece[] pieces; // the array of pieces
	private final String name; // the name of this file
	private final int numOfPieces; // number of pieces constitute the file
	private final Set<Integer> incomplete;
	private String cachedContent;

	public class Piece {
		private final String[] chunksArray;
		private final int index; // piece number
		private int completed;

		/**
		 * Constructs an EMPTY Piece with index i
		 */
		public Piece(int index) {
			assert index >= 0;

			this.index = index;
			this.chunksArray = new String[Constants.NUM_OF_CHUNKS];
			this.completed = 0;
		}

		/**
		 * Constructs a FULL piece with data equal to 'data'
		 * 
		 * @param index -
		 *            the index of this Piece within the File it's a part of
		 * @param content
		 *            the content of this Piece, which may be padded to make the
		 *            Piece the right size. Cannot be <tt>null</tt>
		 */
		public Piece(int index, String content) {
			this(index);

			assert content != null;

			if (content.length() != Constants.PIECE_SIZE)
				content += Manager.filler(Constants.PIECE_SIZE
						- content.length());

			assert content.length() == Constants.PIECE_SIZE;

			int init = 0;
			int fin = Constants.CHUNK_SIZE;
			for (int i = 0; i < Constants.NUM_OF_CHUNKS; i++) {
				put(i, content.substring(init, fin));
				init += Constants.CHUNK_SIZE;
				fin += Constants.CHUNK_SIZE;
			}
		}

		public void put(int chunkIndex, String block) {
			assert 0 <= chunkIndex && chunkIndex < Constants.NUM_OF_CHUNKS;
			assert block.length() == Constants.CHUNK_SIZE;

			if (chunksArray[chunkIndex] == null)
				completed++;

			chunksArray[chunkIndex] = block;

			if (isComplete())
				File.this.incomplete.remove(index);
		}

		public String get(int i) {
			return chunksArray[i];
		}

		public String content() {
			assert isComplete();

			StringBuilder sb = new StringBuilder();
			for (int i = 0; i < Constants.NUM_OF_CHUNKS; i++)
				sb.append(chunksArray[i]);

			return sb.toString();
		}

		public int index() {
			return index;
		}

		public boolean isComplete() {
			return (completed == Constants.NUM_OF_CHUNKS);
		}

		public int completed() {
			return completed;
		}

	}

	/**
	 * @param numOfPieces
	 * @param name
	 * 
	 * @return an empty file named 'name' with 'numOfPieces' total pieces
	 */
	public File(int numOfPieces, String name) {
		this.numOfPieces = numOfPieces;
		this.name = name;
		this.pieces = new Piece[numOfPieces];
		this.incomplete = new HashSet<Integer>();

		for (int i = 0; i < numOfPieces; i++) {
			incomplete.add(i);
			pieces[i] = new Piece(i);
		}
	}

	/**
	 * @param file
	 *            Constructs a File object from the file string
	 */
	public File(String file, String name) {
		this.name = name;
		this.incomplete = Collections.emptySet();

		int size = file.length();
		if (size % Constants.PIECE_SIZE == 0)
			numOfPieces = size / Constants.PIECE_SIZE;
		else
			numOfPieces = size / Constants.PIECE_SIZE + 1;
		pieces = new Piece[numOfPieces];

		// fill in the file with pieces
		int init = 0;
		int fin = Constants.PIECE_SIZE;

		for (int i = 0; i < numOfPieces; i++) {
			String piece = "";
			if (fin >= file.length())
				fin = file.length();
			piece = file.substring(init, fin);
			init += Constants.CHUNK_SIZE;
			fin += Constants.CHUNK_SIZE;
			pieces[i] = new Piece(i, piece);
		}
	}

	private static void populatePiece(Piece piece) {
		// TODO: I don't understand how this works --kuat
		// StringBuffer sb = new StringBuffer();
		// Random r = new Random();
		// for (int i = 0; i < 512; i++) {
		// if (Manager.CHUNKS_OF_LETTERS == null) {
		// System.out.println("asfjawpioefjwaf");
		// }
		// sb.append(Manager.CHUNKS_OF_LETTERS[r
		// .nextInt(Manager.CHUNKS_OF_LETTERS.length)]);
		// }
		// return new Piece(index, sb.toString());
		for (int i = 0; i < Constants.NUM_OF_CHUNKS; i++)
			piece.put(i, Manager.filler(Constants.CHUNK_SIZE));
	}

	/**
	 * Factory method for generating a FULL random file
	 * 
	 * @param name -
	 *            the name of the generated file
	 * @param numPieces -
	 *            the number of pieces contained within this file
	 * @return a File named 'name' containing a total of 'numPieces' pieces
	 */
	public static File generateRandomFile(int numPieces, String name) {
		File retFile = new File(numPieces, name);
		for (int i = 0; i < numPieces; i++) {
			populatePiece(retFile.getPiece(i));
		}
		return retFile;
	}

	/**
	 * Returns a File that contains approximately d times the Pieces, by
	 * randomly selecting Pieces of this to put in the new File
	 * 
	 * @requires 0 <= d <= 1
	 * @param d
	 * @return a File which contains approximately d times the number of Pieces
	 *         in this
	 */
	public File getRandomSubPart(double d) {
		File retFile = new File(this.numOfPieces, this.name);

		Double x = Math.floor(d * this.numOfPieces);
		int newNumOfWholePieces = x.intValue();

		Random r = new Random();
		List<Integer> oneToNumPieces = new ArrayList<Integer>(this.numOfPieces);
		for (int i = 0; i < this.numOfPieces; i++) {
			oneToNumPieces.add(i);
		}

		for (int i = 0; i < newNumOfWholePieces; i++) {
			int rand = r.nextInt(oneToNumPieces.size());
			int y = oneToNumPieces.get(rand);
			oneToNumPieces.remove(rand);
			String pToAdd = pieces[y].content();
			retFile.pieces[y] = retFile.new Piece(y, pToAdd);
		}

		return retFile;
	}

	/**
	 * @return the empty version of this
	 */
	public File getEmptyFile() {
		return new File(this.numOfPieces, this.name);
	}

	/**
	 * 
	 * @return the String content of the file. NULL if the file is not complete.
	 */
	public String content() {
		assert isComplete();

		if (cachedContent == null) {
			StringBuilder sb = new StringBuilder();
			for (int i = 0; i < numOfPieces; i++) {
				String content = pieces[i].content();
				sb.append(content);
			}

			cachedContent = sb.toString();
		}

		return cachedContent;
	}

	/**
	 * 
	 * @param index
	 * @return the piece corresponding to the index. NULL if the file does not
	 *         have the piece.
	 */
	public Piece getPiece(int index) {
		return pieces[index];
	}

	/**
	 * 
	 * @return true if the file is complete. false otherwise.
	 */
	public boolean isComplete() {
		return (incomplete.size() == 0);
	}

	/**
	 * 
	 * @return the name of this File.
	 */
	public String name() {
		return name;
	}

	/**
	 * 
	 * @return the number of pieces this file has
	 */
	public int numberPieces() {
		return numOfPieces;
	}

	@Override
	/**
	 * Simply checks if the two files have the same name, they're as good as
	 * equal for the purposes of simulation
	 */
	public boolean equals(Object o) {
		System.err.println("Don't call equals on files! It is bad!");
		if (o instanceof File) {
			File f = (File) o;
			return this.name().equals(f.name());
		}
		return false;
	}

	public Set<Integer> incomplete() {
		return Collections.unmodifiableSet(incomplete);
	}
}
